library ieee;
use ieee.std_logic_1164.all,ieee.numeric_std.all;

entity alu_16 is
port
(
	sel:in std_logic_vector(4 downto 0); 
	cin:in std_logic;
	a,b:in std_logic_vector(15 downto 0);
	y:out std_logic_vector(15 downto 0);
	zero_flag:out std_logic;
	overflow_flag:out std_logic;
	set_flag:out std_logic;
	cout_flag:out std_logic
);
end alu_16;

architecture alu_16_beh of alu_16 is
procedure logic_to_integer16
(
	in_logic:in std_logic_vector(15 downto 0);
	out_int:out integer
) is
variable count:integer:=0;
variable weight:integer:=1;
begin
	if in_logic(15)='U' then count:=0;
	else
		for I in 0 to 15
		loop
			if in_logic(I)='1' then
				count:=count+weight;
			end if;
			weight:=weight*2;
		end loop;
	end if;
	out_int:=count;
end logic_to_integer16;

procedure integer_to_logic16
(
	in_int : in integer;
	out_logic : out std_logic_vector(15 downto 0)
) is
variable compare:integer:=in_int;
variable weight:integer:=32768;
variable result:std_logic_vector(15 downto 0);
begin
	for I in 15 downto 0
	loop 
		compare:=compare-weight;
		if compare<0 then
			result(I):='0';
			compare:=compare+weight;
		else 
			result(I):='1';
		end if;
		weight:=weight/2;
	end loop;
	out_logic(15 downto 0):=result(15 downto 0);
end integer_to_logic16;
procedure shift_left16
(
	noshift:in std_logic_vector(15 downto 0);
	numshift:in integer;
	shift:out std_logic_vector(15 downto 0)
) is
variable result:std_logic_vector(15 downto 0);
begin
	result:=noshift;
	for J in 1 to numshift
	loop
		for I in 15 downto 1
		loop
			result(I):=result(I-1);
		end loop;
		result(0):='0';
	end loop;
	shift:=result;
end shift_left16;
procedure shift_right16
(
	noshift:in std_logic_vector(15 downto 0);
	numshift:in integer;
	shift:out std_logic_vector(15 downto 0)
) is
variable result:std_logic_vector(15 downto 0);
begin
	result:=noshift;
	for J in 1 to numshift
	loop
		for I in 0 to 14
		loop
			result(I):=result(I+1);
		end loop;
		result(15):='0';
	end loop;
	shift:=result;
end shift_right16;
begin 
	process(sel,a,b,cin)
	variable int_a:integer;
	variable int_b:integer;
	variable int_c:integer;
	variable int_not_b:integer;
	variable selcin:std_logic_vector(2 downto 0);
	variable logicunit,arithunit,noshift:std_logic_vector(15 downto 0);
	variable shift_right_val:std_logic_vector(15 downto 0);
	variable shift_left_val:std_logic_vector(15 downto 0);
	begin
		case sel(1 downto 0)
		is
			when "00"=>logicunit:=A and B;
			when "01"=>logicunit:=A or B;
			when "10"=>logicunit:=A xor B;
			when "11"=>logicunit:=not B;
			when others=>logicunit:=(others=>'X'); 
		end case;
		logic_to_integer16(a,int_a);
		logic_to_integer16(b,int_a);
		selcin:=sel(1 downto 0)&cin;
		logic_to_integer16(a,int_a);
		logic_to_integer16(b,int_b);
		logic_to_integer16(not b,int_not_b);
		case selcin is
			when "000"=>int_c:=int_a;
			when "001"=>int_c:=int_b+1;
			when "010"=>int_c:=int_a+int_b;
			when "011"=>int_c:=int_a+int_b+1;
			when "100"=>int_c:=int_a+int_not_b;
			when "101"=>int_c:=int_a+int_not_b+1;
			when "110"=>int_c:=int_b-1;
			when "111"=>int_c:=int_b;
			when others=>arithunit:=(others=>'X');
		end case;
		if(int_c>65535) then
			cout_flag<='1';
			int_c:=int_c-65536;
		else
			cout_flag<='0';
		end if;
		if(int_a=int_b) then
			zero_flag<='1';
		else
			zero_flag<='0';
		end if;
		if(int_a<32768 and int_b<32768 and int_c>=32768) then
			overflow_flag<='1';
		elsif(int_a>=32768 and int_b>=32768 and int_c<32768) then
			overflow_flag<='1';
		else
			overflow_flag<='0';
		end if;
		if(int_a<32768 and int_b<32768) then
			if(int_a-int_b>=0) then
				set_flag<='0';
			else
				set_flag<='1';
			end if;
		elsif(int_a<32768 and int_b>=32768) then
			set_flag<='0';
		elsif(int_a>=32768 and int_b<32768) then
			set_flag<='1';
		elsif(int_a>=32768 and int_b>=32768) then
			if(int_a-int_b>=0) then
				set_flag<='0';
			else
				set_flag<='1';
			end if;
		end if;
		integer_to_logic16(int_c,arithunit);
		if(sel(2)='1') then
			noshift:=logicunit;
		else 
			noshift:=arithunit;
		end if;
		shift_left16(noshift,int_b,shift_left_val);
		shift_right16(noshift,int_b,shift_right_val);
		case sel(4 downto 3) is
			when "00"=>y<=noshift;
			when "01"=>y<=shift_left_val;
			when "10"=>y<=shift_right_val;
			when "11"=>y<=(others=>'0');
			when others=>y<=(others=>'X');
		end case;
	end process;
end alu_16_beh;
